from .enum import EnumComponent
from .facets.array import ArrayFacetsComponent
from .facets.field import RegularFacetsComponent
from .facets.nested import NestedFacetsComponent
from .facets.object import ObjectFacetsComponent
from .marshmallow import (
    ArrayMarshmallowComponent,
    NumberMarshmallowComponent,
    ObjectMarshmallowComponent,
    RegularMarshmallowComponent,
    StringMarshmallowComponent,
    UIArrayMarshmallowComponent,
    UIMarshmallowComponent,
    UIObjectMarshmallowComponent,
)
from .model import (
    AppModelComponent,
    BlueprintsModelComponent,
    DefaultsModelComponent,
    EDTFIntervalDumperModelComponent,
    ExtResourceModelComponent,
    FacetsModelComponent,
    JSONSchemaModelComponent,
    MappingModelComponent,
    MarshmallowModelComponent,
    PermissionsModelComponent,
    PIDModelComponent,
    ProxyModelComponent,
    RecordDumperModelComponent,
    RecordMetadataModelComponent,
    RecordModelComponent,
    ResourceModelComponent,
    SampleModelComponent,
    SavedModelComponent,
    SearchOptionsModelComponent,
    ServiceModelComponent,
    UIMarshmallowModelComponent,
    UIModelComponent,
)
from .sample import ArraySampleComponent, RegularSampleComponent
from .sort import FieldSortComponent, RegularSortComponent
from .ui import ObjectUIComponent, RegularUIComponent

DEFAULT_COMPONENTS = [
    SearchOptionsModelComponent,
    ArraySampleComponent,
    RegularSampleComponent,
    RegularMarshmallowComponent,
    ObjectMarshmallowComponent,
    RegularUIComponent,
    ObjectUIComponent,
    ArrayMarshmallowComponent,
    UIMarshmallowComponent,
    UIObjectMarshmallowComponent,
    UIArrayMarshmallowComponent,
    EnumComponent,
    AppModelComponent,
    BlueprintsModelComponent,
    FacetsModelComponent,
    JSONSchemaModelComponent,
    MappingModelComponent,
    MarshmallowModelComponent,
    PIDModelComponent,
    PermissionsModelComponent,
    RecordDumperModelComponent,
    RecordMetadataModelComponent,
    RecordModelComponent,
    ResourceModelComponent,
    SampleModelComponent,
    SavedModelComponent,
    ServiceModelComponent,
    UIMarshmallowModelComponent,
    UIModelComponent,
    DefaultsModelComponent,
    ProxyModelComponent,
    ObjectFacetsComponent,
    NestedFacetsComponent,
    RegularFacetsComponent,
    ArrayFacetsComponent,
    ExtResourceModelComponent,
    RegularSortComponent,
    FieldSortComponent,
    StringMarshmallowComponent,
    NumberMarshmallowComponent,
    EDTFIntervalDumperModelComponent,
]



from oarepo_model_builder.schema import ModelSchema
from oarepo_model_builder.utils.python_name import package_name
from oarepo_model_builder.validation import InvalidModelException


def extract_extended_record(included_data, *, context, **kwargs):
    """
    This function extracts the record to be extended. The extended part is always object
        * If it is passed the whole model file, that is, "record" is inside, return that element
        * If there are 'properties' inside, return the whole object
        * As a fallback, return the whole object

    In all cases, it will strip everything apart from "type" and "properties"
    and keep those in "other_properties" inside the context so that other processors
    can add some of those selectively.
    """
    if "record" in included_data:
        extended_object = included_data["record"]
    elif "properties" in included_data:
        extended_object = included_data
    else:
        context["props"] = included_data
        # need to return shallow copy, as there might be a manipulation with the
        # element and context would be corrupted
        return {**included_data}

    context["props"] = extended_object
    return {
        "properties": extended_object.get("properties", {}),
    }


def extend_modify_marshmallow(included_data, *, context, **kwargs):
    """
    This processor modified marshmallow of the extended object. At first, it puts
    marshmallow and ui back to the included data. Then, for the top-level marshmallow & ui.marshmallow
    it converts class -> base-classes and adds import for that.
    For the properties, it marks them as read=False and write=False and for each object, it marks it as
    generate=False - this way, the classes will be reused from the already existing library and not
    generated again.
    """

    def remove_marshmallow_from_children(node):
        ret = {**node}
        node_properties = ret.pop("properties", None)
        node_items = ret.pop("items", None)

        if "marshmallow" in ret:
            convert_marshmallow_class_to_base_class(ret["marshmallow"])
        elif "properties" in ret:
            raise InvalidModelException(
                f"marshmallow section not in object {node}. "
                f"Please pass generated model (records.json5), not the source model."
            )

        if "ui" in ret and "marshmallow" in ret["ui"]:
            convert_marshmallow_class_to_base_class(ret["ui"]["marshmallow"])
        elif "properties" in ret:
            raise InvalidModelException(
                f"ui.marshmallow section not in object {node}. "
                f"Please pass generated model (records.json5), not the source model."
            )

        if node_properties:
            properties = ret.setdefault("properties", {})
            for k, v in node_properties.items():
                remove_marshmallow_from_child(v)
                v = remove_marshmallow_from_children(v)
                properties[k] = v
        if node_items:
            remove_marshmallow_from_child(node.item)
            ret["items"] = remove_marshmallow_from_children(node.item)
        return ret

    def remove_marshmallow_from_child(child):
        # for object/nested, do not set the read & write to False because
        # the extending schema might add more properties.
        # This will generate unnecessary classes, but these might be dealt
        # on later in marshmallow generator
        if "properties" not in child and "items" not in child:
            marshmallow = child.setdefault("marshmallow", {})
            marshmallow.update({"read": False, "write": False})

            ui_marshmallow = child.setdefault("ui", {}).setdefault("marshmallow", {})
            ui_marshmallow.update({"read": False, "write": False})

    def convert_marshmallow_class_to_base_class(marshmallow):
        # pop module & package
        marshmallow.pop("module", None)
        marshmallow.pop("package", None)

        if "class" not in marshmallow:
            return
        clz = marshmallow.pop("class")
        marshmallow.setdefault("base-classes", []).insert(0, clz)
        marshmallow.setdefault("imports", []).append(
            {"import": package_name(clz), "alias": package_name(clz)}
        )

    def as_array(x):
        if isinstance(x, list):
            return x
        if not x:
            return []
        return [x]

    def replace_use_with_extend(data):
        if isinstance(data, dict):
            if ModelSchema.USE_KEYWORD in data:
                data.setdefault(ModelSchema.EXTEND_KEYWORD, []).extend(
                    as_array(data.pop(ModelSchema.USE_KEYWORD))
                )
            if ModelSchema.REF_KEYWORD in data:
                data.setdefault(ModelSchema.EXTEND_KEYWORD, []).extend(
                    as_array(data.pop(ModelSchema.REF_KEYWORD))
                )
            for v in data.values():
                if isinstance(v, (dict, list)):
                    replace_use_with_extend(v)
        elif isinstance(data, list):
            for v in data:
                if isinstance(v, (dict, list)):
                    replace_use_with_extend(v)

    included_data["marshmallow"] = context["props"].get("marshmallow", {})
    included_data["ui"] = context["props"].get("ui", {})
    ret = remove_marshmallow_from_children(included_data)

    for ext in (
        ModelSchema.EXTEND_KEYWORD,
        ModelSchema.REF_KEYWORD,
        ModelSchema.USE_KEYWORD,
    ):
        if ext in context["props"]:
            ret[ext] = context["props"][ext]

    replace_use_with_extend(ret)
    return ret



import marshmallow as ma

from oarepo_model_builder.datatypes import ModelDataType
from oarepo_model_builder.utils.python_name import parent_module
from oarepo_model_builder.validation.extensibility import ExtensibleSchema
from oarepo_model_builder.validation.utils import ImportSchema

from ..ui import ObjectUIComponent, ObjectUIExtraSchema
from .marshmallow import ModelMarshmallowSchema
from .resource import ResourceModelComponent
from .utils import set_default


class JSONSerializerSchema(ma.Schema):
    class Meta:
        unknown = ma.RAISE

    module = ma.fields.String(
        metadata={"doc": "UI serializer class module"},
    )
    class_ = ma.fields.String(
        attribute="class",
        data_key="class",
        metadata={"doc": "UI serializer class qualified name"},
    )
    base_classes = ma.fields.List(
        ma.fields.Str(),
        attribute="base-classes",
        data_key="base-classes",
        metadata={"doc": "List of base classes"},
    )
    extra_code = ma.fields.Str(
        attribute="extra-code",
        data_key="extra-code",
        metadata={
            "doc": "Extra code to be put below the generated ui serializer class"
        },
    )
    list_schema_cls = ma.fields.Str(
        attribute="list-schema-cls",
        data_key="list-schema-cls",
        metadata={"doc": "Marshmallow class for list serialization"},
    )
    format_serializer_cls = ma.fields.Str(
        attribute="format-serializer-cls",
        data_key="format-serializer-cls",
        metadata={"doc": "Class for serializing the resulting json"},
    )
    imports = ma.fields.List(
        ma.fields.Nested(ImportSchema), metadata={"doc": "List of python imports"}
    )
    skip = ma.fields.Boolean()
    generate = ma.fields.Boolean()


class ModelUISchema(ObjectUIExtraSchema):
    marshmallow = ma.fields.Nested(ModelMarshmallowSchema)


class UIModelComponent(ObjectUIComponent):
    eligible_datatypes = [ModelDataType]
    depends_on = [ResourceModelComponent]

    class ModelSchema(ma.Schema):
        ui = ma.fields.Nested(
            ExtensibleSchema("ui.model", ModelUISchema), metadata={"doc": "UI settings"}
        )
        json_serializer = ma.fields.Nested(
            JSONSerializerSchema,
            attribute="json-serializer",
            data_key="json-serializer",
        )

    def before_model_prepare(self, datatype, **kwargs):
        prefix = datatype.definition["module"]["prefix"]
        resources_module = parent_module(datatype.definition["resource"]["module"])

        set_default(datatype, "ui", {})

        json = set_default(datatype, "json-serializer", {})
        json_module = json.setdefault("module", f"{resources_module}.ui")
        json.setdefault("class", f"{json_module}.{prefix}UIJSONSerializer")
        json.setdefault(
            "base-classes", ["oarepo_runtime.resources.LocalizedUIJSONSerializer"]
        )
        json.setdefault(
            "imports",
            [],
        )
        json.setdefault("extra-code", "")
        json.setdefault("generate", True)
        json.setdefault("list_schema_cls", "flask_resources.BaseListSchema")
        json.setdefault(
            "format_serializer_cls", "flask_resources.serializers.JSONSerializer"
        )



from oarepo_model_builder.utils.jinja import split_package_base_name

from ..builders import OutputBuilder
from ..outputs.cfg import CFGOutput


class InvenioRecordResourceSetupCfgBuilder(OutputBuilder):
    TYPE = "invenio_record_resource_setup_cfg"

    def finish(self):
        super().finish()

        output: CFGOutput = self.builder.get_output("cfg", "setup.cfg")

        register_function = split_package_base_name(
            self.current_model.definition["api-blueprint"]["function"]
        )

        output.add_entry_point(
            "invenio_base.api_blueprints",
            self.current_model.definition["api-blueprint"]["alias"],
            f"{register_function[0]}:{register_function[-1]}",
        )

        register_function = split_package_base_name(
            self.current_model.definition["app-blueprint"]["function"]
        )

        output.add_entry_point(
            "invenio_base.blueprints",
            self.current_model.definition["app-blueprint"]["alias"],
            f"{register_function[0]}:{register_function[-1]}",
        )



import copy
import json

from deepdiff import DeepDiff

from ..utils.deepmerge import deepmerge
from ..utils.verbose import log
from . import OutputBase
from .json_stack import JSONStack

try:
    import json5
except ImportError:
    import json as json5


class JSONOutput(OutputBase):
    IGNORE_NODE = JSONStack.IGNORED_NODE
    IGNORE_SUBTREE = JSONStack.IGNORED_SUBTREE
    TYPE = "json"

    def begin(self):
        try:
            with self.builder.filesystem.open(self.path) as f:
                self.original_data = json5.load(f)  # noqa
                self.data = copy.deepcopy(self.original_data)
        except FileNotFoundError:
            self.original_data = None
            self.data = {}
        except ValueError:
            self.original_data = None
            self.data = {}

    @property
    def created(self):
        return self.original_data is None

    def force_clean_output(self):
        self.original_data = None

    def finish(self):
        # create differing but non-empty files
        if DeepDiff(self.data, self.original_data) and (
            self.data or self.original_data
        ):
            self.builder.filesystem.mkdir(self.path.parent)
            log(2, "Saving %s", self.path)
            with self.builder.filesystem.open(self.path, mode="w") as f:
                json.dump(self.data, f, ensure_ascii=False, indent=4)

    def merge(self, value):
        self.data = deepmerge(value, self.data)

    @property
    def modified(self):
        return DeepDiff(self.data, self.original_data) and (
            self.data or self.original_data
        )



from .array import ArrayDataType
from .flatten import FlatObjectDataType
from .nested import NestedDataType
from .object import ObjectDataType

__all__ = ["ObjectDataType", "NestedDataType", "FlatObjectDataType", "ArrayDataType"]



import re

from jinja2 import pass_context

from .python_name import (  # noqa
    Import,
    PythonQualifiedName,
    base_name,
    package_name,
    split_base_name,
    split_package_base_name,
    split_package_name,
)


def sorted_imports(imports):
    # make the entries unique
    imports_dict = {(x["import"], x.get("alias")): x for x in imports}
    imports = list(imports_dict.values())
    # sort them

    imports.sort(key=lambda x: (x["import"], x.get("alias") or ""))
    return imports


def to_import_list(import_object, alias=None, base_classes_name="base-classes"):
    if isinstance(import_object, Import):
        return [{"import": import_object.import_path, "alias": import_object.alias}]
    elif isinstance(import_object, str):
        pn = PythonQualifiedName(import_object)
        if "." in pn.qualified_name:
            return to_import_list(pn.imports)
        else:
            return []
    elif isinstance(import_object, (tuple, list)):
        return [y for x in import_object for y in to_import_list(x)]
    elif isinstance(import_object, dict):
        ret = []
        if "import" in import_object:
            return [import_object]
        if "imports" in import_object:
            ret.extend(to_import_list(import_object["imports"]))
        if base_classes_name in import_object:
            ret.extend(to_import_list(import_object[base_classes_name]))
        if "extra-code" in import_object:
            ret.extend(
                to_import_list(extract_extra_code_imports(import_object["extra-code"]))
            )
        return ret
    return [import_object]


@pass_context
def generate_import(
    ctx, import_object, alias=None, skip=False, base_classes_name="base-classes"
):
    if skip:
        return ""
    import_list = to_import_list(
        import_object, alias=alias, base_classes_name=base_classes_name
    )
    return "\n".join(
        [generate_import_string(ctx, x) for x in sorted_imports(import_list)]
    )


def extract_extra_code_imports(extra_code):
    extra_code = (extra_code or "").strip()
    ret = []
    for match in re.finditer(r"{{(.*?)}}", extra_code):
        ret.extend(PythonQualifiedName(match.groups()[0]).imports)
    return ret


@pass_context
def generate_extra_code_imports(ctx, extra_code):
    return generate_import(ctx, extract_extra_code_imports(extra_code))


def repr_filter(obj):
    if isinstance(obj, str):
        return obj
    return repr(obj)


def generate_extra_code(obj, skip=False):
    if isinstance(obj, dict):
        extra_code = obj.get("extra-code", "")
    else:
        extra_code = obj
    extra_code = (extra_code or "").rstrip()
    if skip or not extra_code:
        return ""

    def replace_classes(matchobj):
        return PythonQualifiedName(matchobj.group(1)).local_name

    return re.sub(r"{{(.*?)}}", replace_classes, extra_code)


def generate_import_string(ctx, import_object):
    import_name = import_object["import"]
    alias = import_object.get("alias")
    ret = []

    if import_name == alias:
        # if import is the same as alias, keep just the import statement
        ret.append(f"import {import_name}")
    else:
        # otherwise break the import to path and base name and alias if defined
        if "." in import_name:
            import_path, import_name = import_name.rsplit(".", maxsplit=1)
        else:
            import_path = None
        if import_path:
            if ctx["current_module"] == import_path:
                return ""  # do not import from the same module
            ret.append(f"from {import_path} import {import_name}")
        else:
            ret.append(f"import {import_name}")
        if alias and alias != import_name:
            ret.append(f"as {alias}")
    return " ".join(ret)


def generate_list(data, separator=", ", start=False, end=False):
    if not data:
        return ""
    ret = []
    if start:
        ret.append(separator)
    for di, d in enumerate(data):
        if di:
            ret.append(separator)
        ret.append(generate_extra_code(d))
    if end:
        ret.append(separator)
    return "".join(ret)


def in_different_package(current_module, value):
    return current_module != package_name(value)


def with_defined_prefix(always_defined_import_prefixes, value):
    return package_name(value) in always_defined_import_prefixes


def class_header(rec, class_name="class", base_classes_name="base-classes"):
    try:
        ret = [
            base_name(rec[class_name]),
        ]
    except KeyError:
        raise KeyError(f'Do not have "{class_name}" inside {rec}')
    base_classes = rec[base_classes_name]
    if base_classes:
        ret.append("(")
        for idx, cls in enumerate(base_classes):
            pn = PythonQualifiedName(cls)
            if idx > 0:
                ret.append(", ")
            ret.append(pn.local_name)
        ret.append(")")
    return "".join(ret)



import marshmallow as ma

from oarepo_model_builder.datatypes import ModelDataType
from oarepo_model_builder.validation.utils import StrictSchema

from ..sample import RegularSampleComponent
from .utils import set_default


class SampleSchema(ma.Schema):
    class Meta:
        unknown = ma.RAISE

    count = ma.fields.Int(metadata={"doc": "Number of generated records"})
    file_ = ma.fields.Str(
        attribute="file",
        data_key="file",
        metadata={"doc": "File into which the records will be generated"},
    )


class SampleModelComponent(RegularSampleComponent):
    eligible_datatypes = [ModelDataType]

    class ModelSchema(StrictSchema):
        sample = ma.fields.Nested(
            SampleSchema, metadata={"doc": "Settings for sample document generator"}
        )

    def before_model_prepare(self, datatype, **kwargs):
        sample = set_default(datatype, "sample", {})
        sample.setdefault("file", "data/sample_data.yaml")



from typing import Dict

from oarepo_model_builder.datatypes import ModelDataType
from oarepo_model_builder.utils.python_name import (
    convert_config_to_qualified_name,
    parent_module,
)

from ..marshmallow import UIObjectMarshmallowComponent
from .defaults import DefaultsModelComponent
from .marshmallow import MarshmallowModelMixin
from .service import ServiceModelComponent
from .utils import set_default


class UIMarshmallowModelComponent(MarshmallowModelMixin, UIObjectMarshmallowComponent):
    eligible_datatypes = [ModelDataType]
    depends_on = [DefaultsModelComponent, ServiceModelComponent]
    model_marshmallow_section = ["ui", "marshmallow"]
    context_registered_classes_name = "ui-marshmallow-classes"
    register_class_names_method = "ui_marshmallow_register_class_name"
    build_class_names_existing_method = "ui_marshmallow_build_class_name_existing"
    build_class_names_new_method = "ui_marshmallow_build_class_name_new"

    def ui_marshmallow_register_class_names(self, *, datatype, classes, **kwargs):
        classes[datatype.definition[self.model_marshmallow_class_name]].append(
            (True, datatype)
        )

    def before_model_prepare(self, datatype, **kwargs):
        prefix = datatype.definition["module"]["prefix"]
        services_module = parent_module(datatype.definition["service"]["module"])

        marshmallow: Dict = set_default(datatype, "ui", "marshmallow", {})
        marshmallow.setdefault("generate", True)
        module = marshmallow.setdefault("module", f"{services_module}.ui_schema")
        marshmallow.setdefault("class", f"{module}.{prefix}UISchema")
        marshmallow.setdefault("extra-code", "")
        marshmallow.setdefault(
            "base-classes", ["oarepo_runtime.services.schema.ui.InvenioUISchema"]
        )
        marshmallow.setdefault("imports", [])
        convert_config_to_qualified_name(marshmallow)

        if "properties" in datatype.definition and "metadata" in (
            datatype.definition["properties"] or {}
        ):
            metadata_marshmallow = set_default(
                datatype, "properties", "metadata", "ui", "marshmallow", {}
            )
            metadata_marshmallow.setdefault("generate", True)
            metadata_module = metadata_marshmallow.setdefault("module", module)
            metadata_marshmallow.setdefault(
                "class", f"{metadata_module}.{prefix}MetadataUISchema"
            )
            metadata_marshmallow.setdefault("extra-code", "")
            metadata_marshmallow.setdefault("base-classes", ["marshmallow.Schema"])
            convert_config_to_qualified_name(metadata_marshmallow)
